# NestJS MongoDB Native Driver Wrapper

## Overview

This package provides a powerful and flexible wrapper around the MongoDB native driver for NestJS. It supports entities, schema generation, and JSON schema validation, allowing developers to work seamlessly with MongoDB while maintaining strong typing and data integrity.

## Features

- **Entity-Based Approach**: Define entities that map directly to MongoDB collections.
- **Schema Generation**: Automatically generate schemas based on entity definitions.
- **JSON Schema Validation**: Ensure data integrity with built-in JSON schema validators.
- **Transaction Support**: Perform complex database operations within ACID-compliant transactions.
- **Index Management**: Create and manage indexes effortlessly.
- **Optimized Queries**: Utilize the native MongoDB driver for maximum performance.

## Installation

```sh
npm install @codemask-labs/nestjs-mongodb
```

```sh
yarn add @codemask-labs/nestjs-mongodb
```

## Getting Started

### 1. Define an Entity

```typescript
import { ObjectId } from 'mongodb';
import { UTCDate } from '@date-fns/utc';
import { Injectable } from '@nestjs/common';
import { Column, Entity, NestedEntity, InjectCollection } from '@codemask-labs/nestjs-mongodb';
import { AuthorizedDevice } from './authorized-device.entity';

const UNIQUE_USER = 'unique_user';

@NestedEntity()
export class AuthorizedDevice {
    @Column()
    readonly fcmToken: string
}

@Entity('users', {
    index: {
        [UNIQUE_USER]: {
            unique: true
        }
    }
})
export class UserEntity {
    @Column()
    readonly name: string;

    @Column()
    readonly surname: string;

    @Column()
    readonly phonenumber: string;

    @Column({ index: UNIQUE_USER })
    readonly email: string;

    @Column()
    readonly password: string;

    @Column()
    readonly age: number;

    @Column({ required: false })
    readonly isVerified?: boolean;

    @Column()
    readonly selectedDevice: AuthorizedDevice;

    @Column({ type: AuthorizedDevice })
    readonly authorizedDevices: Array<AuthorizedDevice>;

    @Column()
    readonly organisation: ObjectId;

    @Column({ type: Date })
    readonly createdAt: UTCDate;

    @Column({ type: Date })
    readonly updatedAt: UTCDate;
}

@Entity('organisations')
export class OrganisationEntity {
    @Column()
    readonly name: string

    @Column({ enum: OrganisationStatus })
    readonly status: OrganisationStatus

    @Column({ type: Date })
    readonly createdAt: UTCDate

    @Column({ type: Date })
    readonly updatedAt: UTCDate
}

@Injectable()
export class UserService {
    constructor(
        @InjectCollection(UserEntity)
        private readonly userCollection: Collection<UserEntity>,
        @InjectCollection(OrganisationEntity)
        private readonly organisationCollection: Collection<OrganisationEntity>
    ) {}

    async createUser() {
        const device: AuthorizedDevice = {
            fcmToken: '<some random fcm token>'
        }

        const organisation = await this.organisationCollection.insertOne({
            document: {
                name: 'Travolta Family Company',
                status: OrganisationStatus.Registered,
                updatedAt: new UTCDate(),
                createdAt: new UTCDate()
            }
        })

        const user = await this.userCollection.insertOne({
            document: {
                age: 71,
                name: 'John',
                surname: 'Travolta',
                phonenumber: '+48123123123',
                email: 'johnny@travoltafamily.com',
                password: 'qwerty12345',
                organisation: organisation.insertedId,
                selectedDevice: device,
                authorizedDevices: [device],
                updatedAt: new UTCDate(),
                createdAt: new UTCDate()
            }
        })

        return {
            createdUserId: user.insertedId
        }
    }
}
```

## Contributing
We welcome contributions! Feel free to open an issue or submit a pull request.

## License
MIT
