"use strict";
Object.defineProperty(exports, "__esModule", { value: true });
exports.getEntityDropIndexNames = exports.getEntityIndexDescriptions = exports.getEntityIndexes = exports.mergeObjectAndArray = exports.getIndexNameFromKey = exports.getEntitySchemaValidator = exports.getEntityValidator = exports.getPropertyValidator = exports.getObjectValidator = exports.getNestedValueTypeValidator = exports.getArrayValidator = exports.getEnumValidator = exports.getNestedEntityValidator = exports.getBsonType = exports.getEntitiesFromMixedList = exports.getDefaultDataSourceOptions = exports.getConnectionString = exports.getEntityMetadata = exports.getCollectionToken = exports.getDataSourceToken = exports.getClientToken = exports.isEmpty = exports.isDataSourceOptions = void 0;
const ramda_1 = require("ramda");
const mongodb_1 = require("mongodb");
const constants_1 = require("./constants");
const metadata_1 = require("./metadata");
const enums_1 = require("./enums");
const isDataSourceOptions = (object) => (0, ramda_1.is)(Object, object) && (0, ramda_1.has)('database', object);
exports.isDataSourceOptions = isDataSourceOptions;
const isEmpty = (value) => (Array.isArray(value) ? value.length === 0 : Object.keys(value).length === 0);
exports.isEmpty = isEmpty;
const getClientToken = (connectionName) => `mongo-client:${connectionName ?? constants_1.DEFAULT_CONNECTION_NAME}`;
exports.getClientToken = getClientToken;
const getDataSourceToken = (connectionName) => `mongo-datasource:${connectionName ?? constants_1.DEFAULT_CONNECTION_NAME}`;
exports.getDataSourceToken = getDataSourceToken;
const getCollectionToken = (entity, connectionName) => `mongo-entity:${connectionName ?? constants_1.DEFAULT_CONNECTION_NAME}:${entity.name}`;
exports.getCollectionToken = getCollectionToken;
const getEntityMetadata = (entity) => metadata_1.Metadata.get(entity, constants_1.ENTITY_METADATA);
exports.getEntityMetadata = getEntityMetadata;
const getConnectionString = (options) => {
    if (options.uri) {
        return options.uri;
    }
    const auth = options.username && options.password ? `${options.username}:${options.password}@` : '';
    const database = options.database ?? '';
    const hosts = options.hosts ?? [
        {
            host: 'localhost',
            port: 27017
        }
    ];
    const nodes = hosts.map(({ host, port }) => `${host}:${port}`).join(',');
    return `mongodb://${auth}${nodes}/${database}`;
};
exports.getConnectionString = getConnectionString;
const getDefaultDataSourceOptions = (options) => (0, ramda_1.mergeDeepWith)(exports.mergeObjectAndArray, constants_1.DEFAULT_MONGODB_MODULE_OPTIONS, options);
exports.getDefaultDataSourceOptions = getDefaultDataSourceOptions;
const getEntitiesFromMixedList = (mixedList = []) => {
    const entities = Array.isArray(mixedList) ? mixedList : Object.values(mixedList ?? {});
    return entities.reduce((acc, entity) => (!metadata_1.Metadata.has(entity, constants_1.ENTITY_WATERMARK) ? acc : [...acc, entity]), []);
};
exports.getEntitiesFromMixedList = getEntitiesFromMixedList;
const getBsonType = (type) => {
    switch (type) {
        case String:
            return enums_1.BSONType.String;
        case Number:
            return enums_1.BSONType.Number;
        case Boolean:
            return enums_1.BSONType.Bool;
        case Date:
            return enums_1.BSONType.Date;
        case Array:
            return enums_1.BSONType.Array;
        case mongodb_1.ObjectId:
            return enums_1.BSONType.ObjectId;
        default:
            return enums_1.BSONType.Object;
    }
};
exports.getBsonType = getBsonType;
const getNestedEntityValidator = (type, options = {}) => {
    const metadata = metadata_1.Metadata.get(type, constants_1.ENTITY_METADATA);
    if (options.array) {
        return {
            bsonType: enums_1.BSONType.Array,
            additionalProperties: false,
            items: (0, exports.getEntityValidator)(metadata.entityColumns)
        };
    }
    return (0, exports.getEntityValidator)(metadata.entityColumns);
};
exports.getNestedEntityValidator = getNestedEntityValidator;
const getEnumValidator = (type, options) => {
    const validator = (0, ramda_1.reject)(exports.isEmpty, {
        enum: Object.values(type)
    });
    if (options.array) {
        return {
            bsonType: enums_1.BSONType.Array,
            additionalProperties: false,
            items: validator
        };
    }
    return validator;
};
exports.getEnumValidator = getEnumValidator;
const getArrayValidator = (type, options = {}) => ({
    bsonType: enums_1.BSONType.Array,
    additionalProperties: false,
    items: (0, ramda_1.reject)(ramda_1.isNil, {
        bsonType: (0, exports.getBsonType)(type),
        uniqueItems: options.unique,
        ...(0, exports.getNestedValueTypeValidator)(options, options.nestedValueTypes)
    })
});
exports.getArrayValidator = getArrayValidator;
const getNestedValueTypeValidator = (options, nestedValueTypes) => {
    if (!nestedValueTypes || !nestedValueTypes.length) {
        return {};
    }
    const [type, ...remainingTypes] = nestedValueTypes;
    const validator = (0, exports.getPropertyValidator)(type);
    if (validator.bsonType === enums_1.BSONType.Array) {
        return (0, ramda_1.reject)(exports.isEmpty, {
            ...validator,
            items: (0, exports.getNestedValueTypeValidator)(options, remainingTypes)
        });
    }
    if (validator.bsonType === enums_1.BSONType.Object) {
        return (0, ramda_1.reject)(exports.isEmpty, {
            ...validator,
            additionalProperties: (0, exports.getNestedValueTypeValidator)(options, remainingTypes)
        });
    }
    return validator;
};
exports.getNestedValueTypeValidator = getNestedValueTypeValidator;
const getObjectValidator = (options = {}) => ({
    bsonType: enums_1.BSONType.Object,
    additionalProperties: (0, ramda_1.reject)(exports.isEmpty, {
        bsonType: enums_1.BSONType.Object,
        properties: {},
        additionalProperties: (0, exports.getNestedValueTypeValidator)(options, options.nestedValueTypes)
    })
});
exports.getObjectValidator = getObjectValidator;
const getPropertyValidator = (type, options = {}) => {
    if (metadata_1.Metadata.has(type, constants_1.NESTED_ENTITY_WATERMARK)) {
        return (0, exports.getNestedEntityValidator)(type, options);
    }
    if (type === Object) {
        return (0, exports.getObjectValidator)(options);
    }
    if (options.enum) {
        return (0, exports.getEnumValidator)(type, options);
    }
    if (options.array) {
        return (0, exports.getArrayValidator)(type, options);
    }
    return (0, ramda_1.reject)(ramda_1.isNil, {
        bsonType: (0, exports.getBsonType)(type)
    });
};
exports.getPropertyValidator = getPropertyValidator;
const getEntityValidator = (columns) => (0, ramda_1.reject)(exports.isEmpty, columns.reduce((acc, column) => {
    if (metadata_1.Metadata.has(column.type, constants_1.ENTITY_WATERMARK)) {
        throw new Error(`Only Nested Entities can be a member of entity. Consider switching type of: ${column.propertyKey} in [class ${column.parent.name}] to ObjectId.`);
    }
    return {
        bsonType: enums_1.BSONType.Object,
        additionalProperties: false,
        required: column.options.required ? [...(acc.required ?? []), column.propertyKey] : acc.required,
        properties: {
            ...acc.properties,
            [column.propertyKey]: (0, exports.getPropertyValidator)(column.type, column.options)
        }
    };
}, {}));
exports.getEntityValidator = getEntityValidator;
const getEntitySchemaValidator = (columns) => ({
    $jsonSchema: (0, exports.getEntityValidator)(columns)
});
exports.getEntitySchemaValidator = getEntitySchemaValidator;
const getIndexNameFromKey = (key) => key.replace(/\.+/g, '_').toUpperCase();
exports.getIndexNameFromKey = getIndexNameFromKey;
const mergeObjectAndArray = (left, right) => {
    if (Array.isArray(left) && Array.isArray(right)) {
        return (0, ramda_1.concat)(left, right);
    }
    if ((0, ramda_1.is)(Object, left) && (0, ramda_1.is)(Object, right)) {
        return (0, exports.mergeObjectAndArray)(left, right);
    }
    return right;
};
exports.mergeObjectAndArray = mergeObjectAndArray;
const getEntityIndexes = (metadata, stackedKeys = []) => {
    const columns = metadata.entityColumns;
    const createIndexOptionMap = metadata.entityOptions.index ?? {};
    return columns.reduce((result, column) => {
        const isNestedEntity = metadata_1.Metadata.has(column.type, constants_1.NESTED_ENTITY_WATERMARK);
        if (isNestedEntity) {
            const nestedEntityMetadata = metadata_1.Metadata.get(column.type, constants_1.ENTITY_METADATA);
            return {
                ...result,
                ...(0, exports.getEntityIndexes)(nestedEntityMetadata, [...stackedKeys, column.propertyKey])
            };
        }
        if ((0, ramda_1.isNil)(column.options.index)) {
            return result;
        }
        const key = [...stackedKeys, column.propertyKey].join('.');
        if ((0, ramda_1.is)(String, column.options.index)) {
            const { index: indexName } = column.options;
            if (!createIndexOptionMap[indexName]) {
                throw new Error(`The index '${indexName}' on property '${column.propertyKey}' in [class ${column.parent.name}] has no index defined. Please add 'index[${indexName}]' in @Entity() or @NestedEntity() options.`);
            }
            return (0, ramda_1.mergeDeepWith)(exports.mergeObjectAndArray, result, {
                [indexName]: {
                    keys: [key],
                    options: createIndexOptionMap[indexName]
                }
            });
        }
        if (Array.isArray(column.options.index)) {
            const { index: indexNames } = column.options;
            indexNames.forEach(indexName => {
                if (!createIndexOptionMap[indexName]) {
                    throw new Error(`The index '${indexName}' on property '${column.propertyKey}' in [class ${column.parent.name}] has no index defined. Please add 'index[${indexName}]' in @Entity() or @NestedEntity() options.`);
                }
            });
            return (0, ramda_1.mergeDeepWith)(exports.mergeObjectAndArray, result, indexNames.reduce((acc, indexName) => (0, ramda_1.mergeDeepWith)(exports.mergeObjectAndArray, acc, {
                [indexName]: {
                    keys: [key],
                    options: createIndexOptionMap[indexName]
                }
            }), {}));
        }
        const indexName = (0, exports.getIndexNameFromKey)(key);
        const { index: createIndexOptions } = column.options;
        return (0, ramda_1.mergeDeepWith)(exports.mergeObjectAndArray, result, {
            [indexName]: {
                keys: [key],
                options: createIndexOptions
            }
        });
    }, {});
};
exports.getEntityIndexes = getEntityIndexes;
const getEntityIndexDescriptions = (entityIndexes) => Object.entries(entityIndexes).map(([name, { keys, options }]) => ({
    name,
    key: keys.reduce((acc, key) => ({
        ...acc,
        [key]: 1
    }), {}),
    ...options
}));
exports.getEntityIndexDescriptions = getEntityIndexDescriptions;
const getEntityDropIndexNames = (currentIndexDescriptions, entityIndexes) => currentIndexDescriptions.reduce((acc, { name, key: currentIndexKeys, ...currentIndexOptions }) => {
    if (!name || name === '_id_') {
        return acc;
    }
    const targetEntityIndex = entityIndexes[name];
    if (!targetEntityIndex) {
        return [...acc, name];
    }
    const differenceInKeys = (0, ramda_1.difference)(Object.keys(currentIndexKeys), targetEntityIndex.keys);
    if (differenceInKeys.length) {
        return [...acc, name];
    }
    const [differenceInOptions] = (0, ramda_1.difference)([currentIndexOptions], [targetEntityIndex.options]);
    if (differenceInOptions && Object.keys(differenceInOptions).some(key => key !== 'v')) {
        return [...acc, name];
    }
    return acc;
}, []);
exports.getEntityDropIndexNames = getEntityDropIndexNames;
//# sourceMappingURL=utils.js.map