import { ListIndexesOptions } from 'mongodb'
import {
    FindOneAndUpsertCommand,
    MongoAggregateCommand,
    MongoCountDocumentsCommand,
    MongoCreateIndexesCommand,
    MongoDropIndexCommand,
    MongoFindCommand,
    MongoFindOneAndUpdateCommand,
    MongoFindOneCommand,
    MongoInsertManyCommand,
    MongoInsertOneCommand,
    MongoUpdateOneCommand,
    MongoUpsertCommand
} from 'lib/types'
import { MongoDocument, MongoCollection, MongoDatabase } from './mongo'
import { CollectionOptions, EntityMetadata } from './types'


export class Collection<T extends MongoDocument> {
    private readonly collection: MongoCollection<T>
    private readonly metadata: EntityMetadata
    private readonly options: CollectionOptions

    constructor(database: MongoDatabase, metadata: EntityMetadata, options?: CollectionOptions) {
        this.collection = database.collection<T>(metadata.collectionName!, options)
        this.metadata = metadata
        this.options = options ?? {}
    }

    /**
     * Finds documents in the collection that match the specified filter.
     * @param command Optional filter criteria and find options.
     * @returns A cursor for the matching documents.
     * @see https://www.mongodb.com/docs/manual/reference/method/db.collection.find/
     */
    find(command?: MongoFindCommand<T>) {
        return this.collection.find(command?.where ?? {}, {
            session: this.options.session,
            ...command?.options
        })
    }

    /**
     * Finds a single document in the collection that matches the specified filter.
     * @param command Optional filter criteria and find options.
     * @returns A promise resolving to the found document or null if no document matches.
     * @see https://www.mongodb.com/docs/manual/reference/method/db.collection.findOne/
     */
    findOne(command?: MongoFindOneCommand<T>) {
        return this.collection.findOne(command?.where ?? {}, {
            session: this.options.session,
            ...command?.options
        })
    }

    /**
     * Finds a single document and updates it, returning either the original or the updated document.
     * @param command Filter criteria, update operations, and options.
     * @returns A promise resolving to the original or updated document, or null if no document matches.
     * @see https://www.mongodb.com/docs/manual/reference/method/db.collection.findOneAndUpdate/
     */
    findOneAndUpdate(command?: MongoFindOneAndUpdateCommand<T>) {
        return this.collection.findOneAndUpdate(command?.where ?? {}, command?.document ?? {}, {
            session: this.options.session,
            ...command?.options
        })
    }

    /**
     * Updates a single document that matches the specified filter.
     * @param command Filter criteria, update operations, and options.
     * @returns A promise resolving to the result of the update operation.
     * @see https://www.mongodb.com/docs/manual/reference/method/db.collection.updateOne/
     */
    updateOne(command?: MongoUpdateOneCommand<T>) {
        return this.collection.updateOne(command?.where ?? {}, command?.document ?? {}, command?.options)
    }

    /**
     * Inserts a single document into the collection.
     * @param command The document to insert and optional settings.
     * @returns A promise resolving to the inserted document.
     * @see https://www.mongodb.com/docs/manual/reference/method/db.collection.insertOne/
     */
    insertOne(command: MongoInsertOneCommand<T>) {
        return this.collection
            .insertOne(command.document, {
                session: this.options.session,
                ...command.options
            })
            .then(result => {
                if (!result.acknowledged) {
                    // todo: something about that?
                }

                return {
                    // eslint-disable-next-line @typescript-eslint/naming-convention
                    _id: result.insertedId,
                    ...command.document
                }
            })
    }

    /**
     * Inserts multiple documents into the collection.
     * @param command The array of documents to insert and optional settings.
     * @returns A promise resolving to the inserted documents.
     * @see https://www.mongodb.com/docs/manual/reference/method/db.collection.insertMany/
     */
    insertMany(command: MongoInsertManyCommand<T>) {
        return this.collection
            .insertMany(command.documents, {
                session: this.options.session,
                ...command.options
            })
            .then(result => {
                if (!result.acknowledged) {
                    // todo: something about that?
                }

                return command.documents.map((insertedDocument, documentIndex) => ({
                    // eslint-disable-next-line @typescript-eslint/naming-convention
                    _id: result.insertedIds[documentIndex],
                    ...insertedDocument
                }))
            })
    }

    /**
     * Updates a single document if it exists, or inserts a new document if it does not (upsert).
     * @param command Filter criteria, update operations, and options.
     * @returns A promise resolving to the result of the upsert operation.
     * @see https://www.mongodb.com/docs/manual/reference/method/db.collection.updateOne/
     */
    upsert(command?: MongoUpsertCommand<T>) {
        return this.collection.updateOne(command?.where ?? {}, command?.document ?? {}, {
            session: this.options.session,
            ...command?.options,
            upsert: true
        })
    }

    /**
     * Finds a document and updates it, or inserts a new document if none matches the filter (findOneAndUpsert).
     * @param command Filter criteria, update operations, and options.
     * @returns A promise resolving to the original or updated document, or null if no document matches.
     * @see https://www.mongodb.com/docs/manual/reference/method/db.collection.findOneAndUpdate/
     */
    findOneAndUpsert(command?: FindOneAndUpsertCommand<T>) {
        return this.collection.findOneAndUpdate(command?.where ?? {}, command?.document ?? {}, {
            returnDocument: 'after',
            session: this.options.session,
            ...command?.options,
            upsert: true
        })
    }

    /**
     * Performs an aggregation framework pipeline against the collection.
     * @param command Pipeline operations and options.
     * @returns An aggregation cursor for the results of the pipeline.
     * @see https://www.mongodb.com/docs/manual/reference/method/db.collection.aggregate/
     */
    aggregate(command?: MongoAggregateCommand<T>) {
        return this.collection.aggregate(command?.pipeline, {
            session: this.options.session,
            ...command?.options
        })
    }

    /**
     * Counts the number of documents in the collection that match the specified filter.
     * @param filter Optional filter criteria.
     * @param options Additional options for the count operation.
     * @returns A promise resolving to the number of matching documents.
     * @see https://www.mongodb.com/docs/manual/reference/method/db.collection.countDocuments/
     */
    countDocuments(command?: MongoCountDocumentsCommand<T>) {
        return this.collection.countDocuments(command?.where, {
            session: this.options.session,
            ...command?.options
        })
    }

    /**
     * Retrieves a list of all indexes on the collection.
     * @param options Optional settings for listing indexes.
     * @returns A promise resolving to an array of index information objects.
     * @see https://www.mongodb.com/docs/manual/reference/method/db.collection.indexes/
     */
    indexes(options?: ListIndexesOptions) {
        return this.collection.indexes(options)
    }

    /**
     * Drops (removes) the specified index from the collection.
     * @param indexName The name of the index to remove.
     * @param options Additional options for the drop operation.
     * @returns A promise resolving to the result of the drop operation.
     * @see https://www.mongodb.com/docs/manual/reference/method/db.collection.dropIndex/
     */
    dropIndex(command: MongoDropIndexCommand) {
        return this.collection.dropIndex(command.indexName, {
            ...command.options,
            session: this.options.session
        })
    }

    /**
     * Creates multiple indexes on the collection based on the provided specifications.
     * @param command The index specifications and optional settings.
     * @returns A promise resolving to the names of the created indexes.
     * @see https://www.mongodb.com/docs/manual/reference/method/db.collection.createIndexes/
     */
    createIndexes(command: MongoCreateIndexesCommand) {
        return this.collection.createIndexes(command.specification, {
            ...command.options,
            session: this.options.session
        })
    }
}
