import { ClientSession, ClientSessionOptions, MongoClient, MongoServerError, ValidateCollectionOptions } from 'mongodb'
import { MongoDatabase, MongoDocument } from './mongo'
import { DataSourceOptions, Class } from './types'
import { Collection } from './collection'
import {
    getEntityMetadata,
    getEntitiesFromMixedList,
    getEntityIndexes,
    getEntityIndexDescriptions,
    getEntitySchemaValidator,
    getEntityDropIndexNames,
    getConnectionString
} from './utils'
import { TransactionManager } from './transaction-manager'

export class DataSource {
    private readonly db: MongoDatabase

    constructor(
        private readonly client: MongoClient,
        private readonly options: DataSourceOptions
    ) {
        this.db = client.db(options.database, options.dbOptions)
    }

    getDatabaseName() {
        return this.db.databaseName
    }

    getCollection<T extends MongoDocument>(entity: Class<T>): Collection<T> {
        const metadata = getEntityMetadata(entity)

        if (!metadata.collectionName) {
            throw new Error(
                `Only entities can have collections. Please make sure your [class ${entity.name}] is properly decorated with @Entity(collectionName: string).`
            )
        }

        return new Collection(this.db, metadata)
    }

    validate<T extends MongoDocument>(entity: Class<T>, options?: ValidateCollectionOptions) {
        const metadata = getEntityMetadata(entity)

        if (!metadata.collectionName) {
            throw new Error(
                `Only entities can have collections. Please make sure your [class ${entity.name}] is properly decorated with @Entity(collectionName: string).`
            )
        }

        return this.db.admin().validateCollection(metadata.collectionName, options)
    }

    /**
     * Starts a session and provides `withTransaction` method to start a transaction with a callback providing `TransactionManager`.
     *
     * **IMPORTANT:** Running operations in parallel is not supported during a transaction. The use of `Promise.all`,
     * `Promise.allSettled`, `Promise.race`, etc to parallelize operations inside a transaction is
     * undefined behaviour.
     *
     * **IMPORTANT:** When running an operation inside a `withTransaction` callback, if it is not
     * provided the explicit session in its options, it will not be part of the transaction and it will not respect timeoutMS.
     *
     *
     * @remarks
     * - If all operations successfully complete and the `commitTransaction` operation is successful, then the provided function will return the result of the provided function.
     * - If the transaction is unable to complete or an error is thrown from within the provided function, then the provided function will throw an error.
     * - If the transaction is manually aborted within the provided function it will not throw.
     * - If the driver needs to attempt to retry the operations, the provided function may be called multiple times.
     *
     * Checkout a descriptive example here:
     * @see https://www.mongodb.com/blog/post/quick-start-nodejs--mongodb--how-to-implement-transactions
     *
     * If a command inside withTransaction fails:
     * - It may cause the transaction on the server to be aborted.
     * - This situation is normally handled transparently by the driver.
     * - However, if the application catches such an error and does not rethrow it, the driver will not be able to determine whether the transaction was aborted or not.
     * - The driver will then retry the transaction indefinitely.
     *
     * To avoid this situation, the application must not silently handle errors within the provided function.
     * If the application needs to handle errors within, it must await all operations such that if an operation is rejected it becomes the rejection of the callback function passed into withTransaction.
     */
    session(options?: ClientSessionOptions) {
        const session = this.client.startSession(options)

        return {
            withTransaction: async <T>(wrapper: (manager: TransactionManager, session: ClientSession) => T) => {
                try {
                    const manager = new TransactionManager(this.db, { session })
                    const result = await wrapper(manager, session)

                    await session.commitTransaction({ timeoutMS: options?.defaultTimeoutMS })

                    return result
                } catch (error) {
                    await session.abortTransaction({ timeoutMS: options?.defaultTimeoutMS })

                    throw error
                }
            }
        }
    }

    async synchronize() {
        if (!this.options.synchronize) {
            return
        }

        const entities = getEntitiesFromMixedList(this.options.entities)

        await Promise.all(
            entities.map(async entity => {
                const metadata = getEntityMetadata(entity)

                if (!metadata.collectionName) {
                    throw new Error(
                        `Missing collection name on entity [class ${entity.name}]. Please make sure it is properly decorated with @Entity(collectionName: string) decorator.`
                    )
                }

                const entityIndexes = getEntityIndexes(metadata)
                const entityIndexDescriptions = getEntityIndexDescriptions(entityIndexes)

                await this.db.createCollection(metadata.collectionName).catch((error: MongoServerError) => {
                    // todo: move code to enum, Collection already exists
                    if (error.code === 48) {
                        return
                    }

                    throw error
                })

                await this.db.command({
                    collMod: metadata.collectionName,
                    validator: getEntitySchemaValidator(metadata.entityColumns)
                })

                if (entityIndexDescriptions.length) {
                    const collection = this.getCollection(entity)
                    const currentIndexDescriptions = await collection.indexes()
                    const dropIndexNames = getEntityDropIndexNames(currentIndexDescriptions, entityIndexes)

                    await Promise.all(dropIndexNames.map(indexName => collection.dropIndex({ indexName })))
                    await collection.createIndexes({
                        specification: entityIndexDescriptions
                    })
                }
            })
        )
    }

    dropDatabase() {
        return this.db.dropDatabase()
    }

    connect() {
        return this.client.connect()
    }

    close(force?: boolean) {
        return this.client.close(force)
    }

    // eslint-disable-next-line @typescript-eslint/member-ordering
    public static createFromOptions(options: DataSourceOptions): DataSource {
        const uri = getConnectionString(options)
        const client = new MongoClient(uri, options.clientOptions)

        return new DataSource(client, options)
    }
}
