import { InjectionToken } from '@nestjs/common'
import { ClientSession, CreateIndexesOptions } from 'mongodb'
import { MongoClientOptions, MongoCollectionOptions, MongoDatabaseOptions } from './mongo'
import { BSONType } from './enums'

// eslint-disable-next-line @typescript-eslint/no-explicit-any
export type Class<T = any> = new (...args: Array<any>) => T

export type MixedList<T> = Array<T> | { [key: string]: T }

export type DataSourceSingleHost = {
    host: string
    port: number
}

export type DataSourceOptions = {
    uri?: string
    hosts?: Array<DataSourceSingleHost>
    database?: string
    username?: string
    password?: string
    synchronize?: boolean
    runMigrations?: boolean
    entities?: MixedList<Class>
    migrations?: MixedList<Class>
    dbOptions?: MongoDatabaseOptions
    clientOptions?: MongoClientOptions
}

export type ForRootOptions = {
    global?: boolean
    connectionName?: string
    options?: DataSourceOptions
}

export type ForRootAsyncOptions = {
    global?: boolean
    connectionName?: string
    inject?: Array<InjectionToken>
    // eslint-disable-next-line @typescript-eslint/no-explicit-any
    useFactory: (...args: Array<any>) => Promise<DataSourceOptions> | DataSourceOptions
}

// eslint-disable-next-line @typescript-eslint/no-explicit-any
export type EntityColumnTypes = StringConstructor | NumberConstructor | BooleanConstructor | ObjectConstructor | Class | Record<string, any>

export type EntityColumn = {
    type: EntityColumnTypes
    parent: Class
    propertyKey: string
    options: EntityColumnOptions
}

export type EntityColumnOptions = {
    /**
     * @description Describes the inner type for the value of this column used for translation to the BSON Type. Available types are Object, Number, Date, Boolean, String or Nested Entity Class.
     */
    type?: EntityColumnTypes

    /**
     * @description A string that describes the schema and has no effect on validation. Starting in MongoDB 5.1, if the description field is specified, MongoDB includes the description in the error output when a document fails validation.
     */
    description?: string

    /**
     * @description Specifies column to be an array of
     */
    array?: boolean

    /**
     * @description If true, each item in the array must be unique. Otherwise, no uniqueness constraint is enforced.
     */
    unique?: boolean

    /**
     * @description Specifies enum values to be used for validation. Required when validation requires an enum.
     */
    enum?: object

    /**
     * @description Requires this column to have some value, if typescript type is optional or null set this value to false.
     * @default false
     */
    required?: boolean

    /**
     * @description Map the index by index name or index names that are defined in @Entity() or @NestedEntity() decorator. Or, use CreateIndexOptions to create an index with options for this column only.
     */
    index?: string | Array<string> | CreateIndexesOptions

    /**
     * @description Describes ordered list of types used to generate nested validators for this column.
     */
    nestedValueTypes?: Array<EntityColumnTypes>
}

export type EntityOptions = {
    index?: Record<string, CreateIndexesOptions>
    validationLevel?: 'off' | 'strict' | 'moderate'
    validationAction?: 'error' | 'warn'
    options?: CollectionOptions
}

export type EntityMetadata = {
    entity: Class
    collectionName: string | null
    entityOptions: EntityOptions
    entityColumns: Array<EntityColumn>
}

export type EntityIndex = {
    keys: Array<string>
    options: CreateIndexesOptions
}

/**
 * @docs https://www.mongodb.com/docs/manual/reference/operator/query/jsonSchema/#available-keywords
 */
export type EntityColumnValidator = {
    /**
     * @description Enumerates the possible JSON types of the field. Available types are "object", "array", "number", "boolean", "string", and "null".
     *
     * MongoDB's implementation of the JSON Schema does not support the "integer" type. Use the bsonType keyword and the "int" or "long" types instead.
     */
    type?: BSONType

    /**
     * @description Accepts same string aliases used for the $type operator.
     */
    bsonType?: BSONType

    description?: string

    /**
     * @description Indicates the minimum length of array.
     */
    minItems?: number

    /**
     * @description Indicates the maximum length of array.
     */
    maxItems?: number

    required?: Array<string>

    enum?: Array<string>

    additionalProperties?: boolean | EntityColumnValidator

    items?: EntityColumnValidator

    /**
     * @description If true, each item in the array must be unique. Otherwise, no uniqueness constraint is enforced.
     */
    uniqueItems?: boolean

    /**
     * Field must match the regular expression.
     */
    pattern?: string

    /**
     * A valid JSON Schema where each value is also a valid JSON Schema object.
     */
    properties?: Record<string, EntityColumnValidator>

    /**
     * In addition to properties requirements, each property name of this object must be a valid regular expression.
     */
    patternProperties?: Record<string, string>
}

export type MongoJsonSchema = {
    $jsonSchema?: EntityColumnValidator
}

export type CollectionOptions = MongoCollectionOptions & {
    session?: ClientSession
}
