import { is, has, reject, isNil, mergeDeepWith, concat, difference } from 'ramda'
import { IndexDescription, IndexDescriptionInfo, IndexDirection, ObjectId } from 'mongodb'
import { DEFAULT_CONNECTION_NAME, DEFAULT_MONGODB_MODULE_OPTIONS, ENTITY_METADATA, ENTITY_WATERMARK, NESTED_ENTITY_WATERMARK } from './constants'
import { Metadata } from './metadata'
import {
    Class,
    DataSourceOptions,
    DataSourceSingleHost,
    EntityColumn,
    EntityColumnOptions,
    EntityColumnTypes,
    EntityColumnValidator,
    EntityIndex,
    EntityMetadata,
    MixedList,
    MongoJsonSchema
} from './types'
import { BSONType } from './enums'

// eslint-disable-next-line @typescript-eslint/no-explicit-any
export const isDataSourceOptions = (object: any): object is DataSourceOptions => is(Object, object) && has('database', object)

// eslint-disable-next-line @typescript-eslint/no-explicit-any, @typescript-eslint/no-unsafe-argument
export const isEmpty = (value: any) => (Array.isArray(value) ? value.length === 0 : Object.keys(value).length === 0)

export const getClientToken = (connectionName?: string) => `mongo-client:${connectionName ?? DEFAULT_CONNECTION_NAME}`

export const getDataSourceToken = (connectionName?: string) => `mongo-datasource:${connectionName ?? DEFAULT_CONNECTION_NAME}`

export const getCollectionToken = (entity: Class, connectionName?: string) =>
    `mongo-entity:${connectionName ?? DEFAULT_CONNECTION_NAME}:${entity.name}`

export const getEntityMetadata = <T>(entity: Class<T>) => Metadata.get<EntityMetadata>(entity, ENTITY_METADATA)

export const getConnectionString = (options: DataSourceOptions): string => {
    if (options.uri) {
        return options.uri
    }

    const auth = options.username && options.password ? `${options.username}:${options.password}@` : ''
    const database = options.database ?? ''
    const hosts: Array<DataSourceSingleHost> = options.hosts ?? [
        {
            host: 'localhost',
            port: 27017
        }
    ]

    const nodes = hosts.map(({ host, port }) => `${host}:${port}`).join(',')

    return `mongodb://${auth}${nodes}/${database}`
}

export const getDefaultDataSourceOptions = (options: DataSourceOptions): DataSourceOptions =>
    // eslint-disable-next-line @typescript-eslint/no-unsafe-return
    mergeDeepWith(mergeObjectAndArray, DEFAULT_MONGODB_MODULE_OPTIONS, options)

export const getEntitiesFromMixedList = (mixedList: MixedList<Class> = []): Array<Class> => {
    const entities = Array.isArray(mixedList) ? mixedList : Object.values(mixedList ?? {})

    return entities.reduce((acc, entity) => (!Metadata.has(entity, ENTITY_WATERMARK) ? acc : [...acc, entity]), [] as Array<Class>)
}

export const getBsonType = (type: EntityColumnTypes) => {
    switch (type) {
        case String:
            return BSONType.String
        case Number:
            return BSONType.Number
        case Boolean:
            return BSONType.Bool
        case Date:
            return BSONType.Date
        case Array:
            return BSONType.Array
        case ObjectId:
            return BSONType.ObjectId
        default:
            return BSONType.Object
    }
}

export const getNestedEntityValidator = (type: Class, options: EntityColumnOptions = {}): EntityColumnValidator => {
    const metadata = Metadata.get<EntityMetadata>(type, ENTITY_METADATA)

    if (options.array) {
        return {
            bsonType: BSONType.Array,
            additionalProperties: false,
            items: getEntityValidator(metadata.entityColumns)
        }
    }

    return getEntityValidator(metadata.entityColumns)
}

export const getEnumValidator = (type: object, options: EntityColumnOptions): EntityColumnValidator => {
    const validator: EntityColumnValidator = reject(isEmpty, {
        enum: Object.values(type)
    })

    if (options.array) {
        return {
            bsonType: BSONType.Array,
            additionalProperties: false,
            items: validator
        }
    }

    return validator
}

export const getArrayValidator = (type: EntityColumnTypes, options: EntityColumnOptions = {}): EntityColumnValidator => ({
    bsonType: BSONType.Array,
    additionalProperties: false,
    items: reject(isNil, {
        bsonType: getBsonType(type),
        uniqueItems: options.unique,
        ...getNestedValueTypeValidator(options, options.nestedValueTypes)
    })
})

export const getNestedValueTypeValidator = (options: EntityColumnOptions, nestedValueTypes?: Array<EntityColumnTypes>): EntityColumnValidator => {
    if (!nestedValueTypes || !nestedValueTypes.length) {
        return {}
    }

    const [type, ...remainingTypes] = nestedValueTypes
    const validator = getPropertyValidator(type)

    if (validator.bsonType === BSONType.Array) {
        return reject(isEmpty, {
            ...validator,
            items: getNestedValueTypeValidator(options, remainingTypes)
        })
    }

    if (validator.bsonType === BSONType.Object) {
        return reject(isEmpty, {
            ...validator,
            additionalProperties: getNestedValueTypeValidator(options, remainingTypes)
        })
    }

    return validator
}

export const getObjectValidator = (options: EntityColumnOptions = {}): EntityColumnValidator => ({
    bsonType: BSONType.Object,
    additionalProperties: reject(isEmpty, {
        bsonType: BSONType.Object,
        properties: {},
        additionalProperties: getNestedValueTypeValidator(options, options.nestedValueTypes)
    })
})

export const getPropertyValidator = (type: EntityColumnTypes, options: EntityColumnOptions = {}): EntityColumnValidator => {
    if (Metadata.has(type, NESTED_ENTITY_WATERMARK)) {
        return getNestedEntityValidator(type, options)
    }

    if (type === Object) {
        return getObjectValidator(options)
    }

    if (options.enum) {
        return getEnumValidator(type, options)
    }

    if (options.array) {
        return getArrayValidator(type, options)
    }

    return reject(isNil, {
        bsonType: getBsonType(type)
    })
}

export const getEntityValidator = (columns: Array<EntityColumn>) =>
    reject(
        isEmpty,
        columns.reduce<EntityColumnValidator>((acc, column) => {
            if (Metadata.has(column.type, ENTITY_WATERMARK)) {
                throw new Error(
                    `Only Nested Entities can be a member of entity. Consider switching type of: ${column.propertyKey} in [class ${column.parent.name}] to ObjectId.`
                )
            }

            return {
                bsonType: BSONType.Object,
                additionalProperties: false,
                required: column.options.required ? [...(acc.required ?? []), column.propertyKey] : acc.required,
                properties: {
                    ...acc.properties,
                    [column.propertyKey]: getPropertyValidator(column.type, column.options)
                }
            }
        }, {})
    )

export const getEntitySchemaValidator = (columns: Array<EntityColumn>): MongoJsonSchema => ({
    $jsonSchema: getEntityValidator(columns)
})

export const getIndexNameFromKey = (key: string) => key.replace(/\.+/g, '_').toUpperCase()

// eslint-disable-next-line @typescript-eslint/no-explicit-any
export const mergeObjectAndArray = (left: any, right: any): any => {
    if (Array.isArray(left) && Array.isArray(right)) {
        return concat(left, right)
    }

    if (is(Object, left) && is(Object, right)) {
        return mergeObjectAndArray(left, right)
    }

    return right
}

export const getEntityIndexes = (metadata: EntityMetadata, stackedKeys: Array<string> = []): Record<string, EntityIndex> => {
    const columns = metadata.entityColumns
    const createIndexOptionMap = metadata.entityOptions.index ?? {}

    return columns.reduce(
        (result, column) => {
            const isNestedEntity = Metadata.has(column.type, NESTED_ENTITY_WATERMARK)

            if (isNestedEntity) {
                const nestedEntityMetadata = Metadata.get<EntityMetadata>(column.type, ENTITY_METADATA)

                return {
                    ...result,
                    ...getEntityIndexes(nestedEntityMetadata, [...stackedKeys, column.propertyKey])
                }
            }

            if (isNil(column.options.index)) {
                return result
            }

            const key = [...stackedKeys, column.propertyKey].join('.')

            if (is(String, column.options.index)) {
                const { index: indexName } = column.options

                if (!createIndexOptionMap[indexName]) {
                    throw new Error(
                        `The index '${indexName}' on property '${column.propertyKey}' in [class ${column.parent.name}] has no index defined. Please add 'index[${indexName}]' in @Entity() or @NestedEntity() options.`
                    )
                }

                // eslint-disable-next-line @typescript-eslint/no-unsafe-return
                return mergeDeepWith(mergeObjectAndArray, result, {
                    [indexName]: {
                        keys: [key],
                        options: createIndexOptionMap[indexName]
                    }
                })
            }

            if (Array.isArray(column.options.index)) {
                const { index: indexNames } = column.options

                indexNames.forEach(indexName => {
                    if (!createIndexOptionMap[indexName]) {
                        throw new Error(
                            `The index '${indexName}' on property '${column.propertyKey}' in [class ${column.parent.name}] has no index defined. Please add 'index[${indexName}]' in @Entity() or @NestedEntity() options.`
                        )
                    }
                })

                // eslint-disable-next-line @typescript-eslint/no-unsafe-return
                return mergeDeepWith(
                    mergeObjectAndArray,
                    result,
                    indexNames.reduce(
                        (acc, indexName) =>
                            // eslint-disable-next-line @typescript-eslint/no-unsafe-return
                            mergeDeepWith(mergeObjectAndArray, acc, {
                                [indexName]: {
                                    keys: [key],
                                    options: createIndexOptionMap[indexName]
                                }
                            }),
                        {} as Record<string, EntityIndex>
                    )
                )
            }

            const indexName = getIndexNameFromKey(key)
            const { index: createIndexOptions } = column.options

            // eslint-disable-next-line @typescript-eslint/no-unsafe-return
            return mergeDeepWith(mergeObjectAndArray, result, {
                [indexName]: {
                    keys: [key],
                    options: createIndexOptions
                }
            })
        },
        {} as Record<string, EntityIndex>
    )
}

export const getEntityIndexDescriptions = (entityIndexes: Record<string, EntityIndex>): Array<IndexDescription> =>
    Object.entries(entityIndexes).map(([name, { keys, options }]) => ({
        name,
        key: keys.reduce(
            (acc, key) => ({
                ...acc,
                [key]: 1 // todo: add support for key direction to be specified on @Column() generator or @Entity() or @NestedEntity()
            }),
            {} as Record<string, IndexDirection>
        ),
        ...options
    }))

export const getEntityDropIndexNames = (currentIndexDescriptions: Array<IndexDescriptionInfo>, entityIndexes: Record<string, EntityIndex>) =>
    currentIndexDescriptions.reduce((acc, { name, key: currentIndexKeys, ...currentIndexOptions }) => {
        if (!name || name === '_id_') {
            return acc
        }

        const targetEntityIndex = entityIndexes[name]

        if (!targetEntityIndex) {
            return [...acc, name]
        }

        const differenceInKeys = difference(Object.keys(currentIndexKeys), targetEntityIndex.keys)

        if (differenceInKeys.length) {
            return [...acc, name]
        }

        const [differenceInOptions] = difference([currentIndexOptions], [targetEntityIndex.options])

        if (differenceInOptions && Object.keys(differenceInOptions).some(key => key !== 'v')) {
            return [...acc, name]
        }

        return acc
    }, [] as Array<string>)
