import { DynamicModule, FactoryProvider, Module } from '@nestjs/common'
import { MongoClient } from 'mongodb'
import { randomUUID } from 'node:crypto'
import {
    ForRootOptions,
    getDefaultDataSourceOptions,
    getConnectionString,
    getClientToken,
    getDataSourceToken,
    DataSource,
    INTERNAL_DATA_SOURCE,
    ForRootAsyncOptions,
    DataSourceOptions,
    Class,
    getCollectionToken
} from 'lib/common'
import { MongodbService } from './mongodb.service'
import { MongodbTestService } from './mongodb-test.service'

@Module({})
export class MongodbModule {
    static forRoot({ global, connectionName, options = {} }: ForRootOptions): DynamicModule {
        const defaultOptions = getDefaultDataSourceOptions(options)
        const uri = getConnectionString(defaultOptions)
        const clientToken = getClientToken(connectionName)
        const dataSourceToken = getDataSourceToken(connectionName)

        return {
            global: global ?? true,
            module: MongodbModule,
            exports: [clientToken, dataSourceToken],
            providers: [
                {
                    provide: clientToken,
                    useFactory: () => new MongoClient(uri, defaultOptions.clientOptions)
                },
                {
                    provide: dataSourceToken,
                    inject: [clientToken],
                    useFactory: (client: MongoClient) => new DataSource(client, defaultOptions)
                },
                {
                    provide: INTERNAL_DATA_SOURCE,
                    inject: [dataSourceToken],
                    useFactory: (dataSource: DataSource) => dataSource
                },
                {
                    provide: randomUUID(),
                    useClass: MongodbService
                }
            ]
        }
    }

    static forRootAsync({ global, connectionName, inject, useFactory }: ForRootAsyncOptions): DynamicModule {
        const clientToken = getClientToken(connectionName)
        const dataSourceToken = getDataSourceToken(connectionName)
        const dataSourceOptionsToken = randomUUID()

        return {
            global: global ?? true,
            module: MongodbModule,
            exports: [clientToken, dataSourceToken],
            providers: [
                {
                    provide: dataSourceOptionsToken,
                    inject,
                    useFactory
                },
                {
                    provide: clientToken,
                    inject: [dataSourceOptionsToken],
                    useFactory: (options: DataSourceOptions) => {
                        const defaultOptions = getDefaultDataSourceOptions(options)
                        const uri = getConnectionString(defaultOptions)

                        return new MongoClient(uri, defaultOptions.clientOptions)
                    }
                },
                {
                    provide: dataSourceToken,
                    inject: [clientToken, dataSourceOptionsToken],
                    useFactory: (client: MongoClient, options: DataSourceOptions) => {
                        const defaultOptions = getDefaultDataSourceOptions(options)

                        return new DataSource(client, defaultOptions)
                    }
                },
                {
                    provide: INTERNAL_DATA_SOURCE,
                    inject: [dataSourceToken],
                    useFactory: (dataSource: DataSource) => dataSource
                },
                {
                    provide: randomUUID(),
                    useClass: MongodbService
                }
            ]
        }
    }

    static forTestAsync({ global, connectionName, inject, useFactory }: ForRootAsyncOptions): DynamicModule {
        const clientToken = getClientToken(connectionName)
        const dataSourceToken = getDataSourceToken(connectionName)
        const dataSourceOptionsToken = randomUUID()

        return {
            global: global ?? true,
            module: MongodbModule,
            exports: [clientToken, dataSourceToken],
            providers: [
                {
                    provide: dataSourceOptionsToken,
                    inject,
                    // eslint-disable-next-line @typescript-eslint/no-explicit-any, functional/functional-parameters
                    useFactory: async (...args: Array<any>) => {
                        // eslint-disable-next-line @typescript-eslint/no-unsafe-argument
                        const dataSourceOptions = await useFactory(...args)

                        return { ...dataSourceOptions, database: randomUUID() }
                    }
                },
                {
                    provide: clientToken,
                    inject: [dataSourceOptionsToken],
                    useFactory: (options: DataSourceOptions) => {
                        const defaultOptions = getDefaultDataSourceOptions(options)
                        const uri = getConnectionString(defaultOptions)

                        return new MongoClient(uri, defaultOptions.clientOptions)
                    }
                },
                {
                    provide: dataSourceToken,
                    inject: [clientToken, dataSourceOptionsToken],
                    useFactory: (client: MongoClient, options: DataSourceOptions) => {
                        const defaultOptions = getDefaultDataSourceOptions(options)

                        return new DataSource(client, defaultOptions)
                    }
                },
                {
                    provide: INTERNAL_DATA_SOURCE,
                    inject: [dataSourceToken],
                    useFactory: (dataSource: DataSource) => dataSource
                },
                {
                    provide: randomUUID(),
                    useClass: MongodbTestService
                }
            ]
        }
    }

    static forFeature(entities: Array<Class>, connectionName?: string): DynamicModule {
        const dataSourceToken = getDataSourceToken(connectionName)
        const providers: Array<FactoryProvider> = entities.map(entity => ({
            provide: getCollectionToken(entity, connectionName),
            inject: [dataSourceToken],
            useFactory: (dataSource: DataSource) => dataSource.getCollection(entity)
        }))

        return {
            module: MongodbModule,
            providers,
            exports: providers.map(({ provide }) => provide)
        }
    }

    static forTest(): DynamicModule {
        return {
            module: MongodbModule,
            providers: [],
            exports: []
        }
    }
}
